-- # import eq
-- # import nat
-- # import vec
-- # import thunk
-- # import order
-- # import cumSum
-- # import lazyVec
-- # import thunkedVec
-- # import bankersQueueInvariants

data BankersQueue (a : Set) : (Δ, L : ℕ) -> Set where
  BQ
    :  (Δ, L : ℕ)
    -> (Cs : Vec (L + Δ) ℕ)
    -> bqInvariant 2 (L + Δ) (2 * Δ) Cs
    -> LazyVec a (L + Δ) Cs
    -> Vec L a
    -> BankersQueue a Δ L

empty (a : Set) : BankersQueue a 0 0
empty = BQ 0 0 [] (bqNil 2 0) lNil []

singletonQ (a : Set) : a -> BankersQueue a 1 0
singletonQ x =
  BQ 1 0 [0] (bqSingleton 2 2) (lCons x $ return lNil) []

snocL' : (Δ, L : ℕ) -> ℕ
snocL' 0       L = 0
snocL' (suc _) L = L + 1

snocΔ' : (Δ, L : ℕ) -> ℕ
snocΔ' 0       L = 2 * L + 1
snocΔ' (suc Δ) L = Δ

-- I pretend that the arithmetic operations below take constant time,
-- and do not include them in the analysis. For the unary numbers
-- which I am using this is not true, but it should not be hard to
-- work around this with proper support for machine integers.

(▹) (Δ, L : ℕ) (a : Set)
  :  a
  -> BankersQueue a Δ L
  -> Thunk 2 (BankersQueue a (snocΔ' Δ L) (snocL' Δ L))
x ▹ BQ 0 0 _ _ lNil [] = return' $ singletonQ x
x ▹ BQ (suc Δ) L Cs inv f r = thunkM.do
  f' <- payLazyVec 2 f
  return $ BQ Δ (L + 1) (Cs ⊖ 2) (bqSub {2} 2 inv) f' (x :: r)
x ▹ BQ 0 (suc L) Cs inv f r = thunkM.do
    f' <- append' f (fromVec =<< tReverse (x :: r))
    return' $ BQ (2 * L + 3) 0 Cs' inv' f' []
  where
  Cs' : Vec (2 * L + 3) ℕ
  Cs' = map suc (init Cs) ++ (L + 3) :: replicate {L + 2} 1

  inv' : bqInvariant 2 (2 * L + 3) (2 * (2 * L + 3)) Cs'
  inv' = bqAppendRev' {0} inv

viewLL' : (Δ, L : ℕ) -> ℕ
viewLL' 0       L = 0
viewLL' (suc _) L = L

viewLΔ' (Δ, L : ℕ) : Δ ≥ 1 ∔ L ≥ 1 -> ℕ
viewLΔ' {0}     {0}     (left ge)  = absurdLE ge
viewLΔ' {0}     {0}     (right ge) = absurdLE ge
viewLΔ' {0}     {suc L} nonEmpty   = 2 * L + 1
viewLΔ' {suc Δ} {L}     nonEmpty   = Δ

viewL (Δ, L : ℕ) (a : Set)
  :  BankersQueue a Δ L
  -> (nonEmpty : Δ ≥ 1 ∔ L ≥ 1)
  -> Thunk 4 (a × BankersQueue a (viewLΔ' nonEmpty) (viewLL' Δ L))
viewL (BQ 0 0 _ _ _ _) (left ge)  = absurdLE ge
viewL (BQ 0 0 _ _ _ _) (right ge) = absurdLE ge
viewL (BQ 0 1 _ _ (lCons x _) [y]) _ = return' $
  pair x $ singletonQ y
viewL (BQ (suc Δ) L (C :: Cs) inv f r) _ = thunkM.do
    f¹ <- payLazyVec 3 f
    f² <- slower prop $ lTail f¹
    return' $ pair (lHead f¹) $
      BQ Δ L (tail ((C :: Cs) ⊖ 3))
         (bqRaiseΔ {2} (dropPlusMinus $ leS leZ) $ bqTail {2} inv)
         f² r
  where
  prop : C ∸ 3 ≤ 0
  prop = liftEq $ subtractLarger $
    inv 1 (leS leZ) `transitive` plusLeftMon {2} minLeft
viewL (BQ 0 (suc (suc L)) (C :: Cs) inv f r) _ = thunkM.do
    f¹ <- payLazyVec 3 f
    f² <- slower prop₁ $ lTail f¹
    f³ <- append' f² (fromVec =<< tReverse r)
    return $ pair (lHead f¹) $ BQ (2 * L + 3) 0 Cs' inv' f³ []
  where
  prop₁ : C ∸ 3 ≤ 0
  prop₁ = smallerRelated `transitive` inv 1 (leS leZ)

  Cs' : Vec (2 * L + 3) ℕ
  Cs' = map suc (init $ Cs ⊖ (3 ∸ C)) ++ (L + 3) :: replicate {L + 2} 1

  inv' : bqInvariant 2 (2 * L + 3) (2 * (2 * L + 3)) Cs'
  inv' = bqAppendRev' {0} $ bqTail {2} {_} {0} inv
