-- # import nat
-- # import order
-- # import pair

data between : ℕ -> ℕ -> ℕ -> Set where
  bZZ (i : ℕ) : between 0 0 i
  bZS (i, j : ℕ)
    : between 0 i j -> between 0 (i + 1) (j + 1)
  bSS (i, j, k : ℕ)
    : between i j k -> between (i + 1) (j + 1) (k + 1)

toLE (i, j, k : ℕ)
  : between i j k -> i ≤ j × j ≤ k
toLE bZZ = pair leZ leZ
toLE (bZS b) with toLE b =>
  _ _ (pair _ le) = pair leZ (leS le)
toLE (bSS b) with toLE b =>
  _ _ (pair le le') = pair (leS le) (leS le')

fromLE (i, j, k : ℕ)
  : i ≤ j -> j ≤ k -> between i j k
fromLE leZ leZ            = bZZ
fromLE leZ (leS le)       = bZS (fromLE leZ le)
fromLE (leS le) (leS le') = bSS (fromLE le le')

betweenPlusPlus (i, j, k : ℕ) : between i (i + j) (i + j + k)
betweenPlusPlus = fromLE plusIncreasingRight plusIncreasingRight

limitBetweenLeft (i, j, k : ℕ)
  : (i' : ℕ) -> between (i + i') j k -> between i j k
limitBetweenLeft _ b with toLE b =>
  _ _ _ (pair le le') = fromLE (plusIncreasingRight `transitive` le) le'

limitBetweenRight (i, j, k : ℕ)
  : (k' : ℕ) -> between i j k -> between i j (k + k')
limitBetweenRight _ b with toLE b =>
  _ _ _ (pair le le') = fromLE le (le' `transitive` plusIncreasingRight)

betweenMon (i, j, k, i', j', k' : ℕ)
  :  between i  j  k
  -> between i' j' k'
  -> between (i + i') (j + j') (k + k')
betweenMon b₁ b₂ with toLE b₁, toLE b₂ =>
  _ _ _ (pair le₁ le₁') (pair le₂ le₂') =
    fromLE (plusMon le₁ le₂) (plusMon le₁' le₂')

betweenMonSubtract (i, j, k, i', j', k' : ℕ)
  :  between i  j  k
  -> between i' j' k'
  -> between (i ∸ k') (j ∸ j') (k ∸ i')
betweenMonSubtract b₁ b₂ with toLE b₁, toLE b₂ =>
  _ _ _ (pair le₁ le₁') (pair le₂ le₂') =
    fromLE (subtractMonotoneAntimonotone le₁  le₂')
           (subtractMonotoneAntimonotone le₁' le₂)

splitBetween (j, i, k : ℕ)
  : (n : ℕ) -> between i j k -> between i j n ∔ between (n + 1) j k
splitBetween {j} n b with toLE b, j ≤?> n =>
  _ _ _ (pair le le') (left  jLeN) = left  $ fromLE le   jLeN
  _ _ _ (pair le le') (right nLeJ) = right $ fromLE nLeJ le'

------------------------------------------------------------------------
-- Custom betweens

b₁₁₁ : between 1 1 1
b₁₁₁ = bSS bZZ
