-- # import nat
-- # import vec
-- # import eq
-- # import thunk
-- # import cumSum
-- # import thunkedOrder

data LazyVec (a : Set) : (ℓ : ℕ) -> (costs : Vec ℓ ℕ) -> Set where
  lNil : LazyVec a 0 []
  lCons (ℓ, C : ℕ) (Cs : Vec ℓ ℕ)
    : a -> Thunk C (LazyVec a ℓ Cs) -> LazyVec a (ℓ + 1) (C :: Cs)

lHead (a : Set) (ℓ : ℕ) (Cs : Vec (ℓ + 1) ℕ)
  : LazyVec a (ℓ + 1) Cs -> a
lHead (lCons x _) = x

lTail (a : Set) (ℓ : ℕ) (Cs : Vec (ℓ + 1) ℕ)
  : LazyVec a (ℓ + 1) Cs -> Thunk (head Cs) (LazyVec a ℓ (tail Cs))
lTail (lCons _ f) = f

------------------------------------------------------------------------

-- These pay functions are superseded by the generalised pay in the
-- new library. Consider these functions as being implemented like the
-- ones in Agda2/LazyVec/General.agda. (This requires replacing the
-- data type LazyVec with an isomorphic type defined using a
-- function.)

pay₁LazyVec (a : Set) (ℓ : ℕ) (Cs : Vec ℓ ℕ)
  : LazyVec a ℓ Cs -> Thunk 1 (LazyVec a ℓ (Cs ⊖ 1))
pay₁LazyVec lNil                    = return' lNil
pay₁LazyVec (lCons {_} {suc C} x f) = return (lCons x) `ap` pay 1 f
pay₁LazyVec (lCons {_} {0}     x f) = {- step $ -} thunkM.do
  f' <- pay₁LazyVec =<< f
  return (lCons x $ return f')

payLazyVec (ℓ : ℕ) (Cs : Vec ℓ ℕ) (a : Set)
  : (n : ℕ) -> LazyVec a ℓ Cs -> Thunk n (LazyVec a ℓ (Cs ⊖ n))
payLazyVec 0       xs = return xs
payLazyVec (suc n) xs = {- step $ -} thunkM.do
  xs' <- pay₁LazyVec xs
  substThunk refl (cong'' (LazyVec _ _) $ prop_sub_sub_plus _ 1 n) $
    payLazyVec n xs'

------------------------------------------------------------------------

-- rotate f r a ≈ f ++ reverse r ++ a

rotate (ℓ₁, ℓ₂ : ℕ) (Cs₁ : Vec ℓ₁ ℕ) (Cs₂ : Vec ℓ₂ ℕ) (a : Set)
  :  LazyVec a ℓ₁ Cs₁ -> Vec (ℓ₁ + 1) a -> LazyVec a ℓ₂ Cs₂
  -> Thunk 1 (LazyVec a (ℓ₁ + (ℓ₁ + 1 + ℓ₂))
                         (map suc Cs₁ ++ replicate {ℓ₁ + 1} 0 ++ Cs₂))
rotate lNil        [y]      a = return' $ lCons y $ return a
rotate {suc ℓ₁} {_} {_} {Cs₂}
       (lCons x f) (y :: r) a = step $
  rewrite prop_replicate_cons {ℓ₁} {_} {_} {0} {Cs₂} in
    return $ lCons x $ thunkM.do
      f' <- f
      rotate f' r (lCons y $ return a)

fromVec (ℓ : ℕ) (a : Set)
  : Vec ℓ a -> Thunk 1 (LazyVec a ℓ (replicate 1))
fromVec []        = return' lNil
fromVec (x :: xs) = step $ return $ lCons x (fromVec xs)

slowerLazyVec (C', C, ℓ : ℕ) (a : Set)
  :  LazyVec a ℓ (replicate C)
  -> Thunk (C' + 1) (LazyVec a ℓ (replicate (C + (C' + 1))))
slowerLazyVec lNil         = return' lNil
slowerLazyVec (lCons x xs) = step $
  return' $ lCons x (slowerLazyVec =<< xs)

append (ℓ₁, ℓ₂ : ℕ) (Cs₁ : Vec ℓ₁ ℕ) (Cs₂ : Vec ℓ₂ ℕ) (a : Set)
  :  LazyVec a ℓ₁ Cs₁ -> LazyVec a ℓ₂ Cs₂
  -> Thunk 1 (LazyVec a (ℓ₁ + ℓ₂) (map suc Cs₁ ++ Cs₂))
append lNil        ys = return' ys
append (lCons x f) ys = step $ return $ lCons x $ thunkM.do
  f' <- f
  append f' ys

append' (ℓ₁ : ℕ) (Cs₁ : Vec (ℓ₁ + 1) ℕ)
        (ℓ₂, m : ℕ) (Cs₂ : Vec ℓ₂ ℕ) (a : Set)
  :  LazyVec a (ℓ₁ + 1) Cs₁ -> Thunk m (LazyVec a ℓ₂ Cs₂)
  -> Thunk 1 (LazyVec a (ℓ₁ + ℓ₂ + 1) (map suc (init Cs₁) ++ m :: Cs₂))
append' {_} {[_]}         (lCons x _) ys = return' $ lCons x ys
append' {_} {_ :: _ :: _} (lCons x f) ys = step $
  return $ lCons x $ thunkM.do
    f' <- f
    append' f' ys

-- We cheat below (treating ≤?> as a constant-time operation) in order
-- to be able to compare with other implementations of lazy
-- insertion-sort.

insert (C, ℓ : ℕ)
  :  ℕ -> LazyVec ℕ ℓ (replicate C)
  -> Thunk 2 (LazyVec ℕ (ℓ + 1) (replicate (C + 2)))
insert = insert'
  where
  mutual
    insert' (C, ℓ : ℕ)
      :  ℕ -> LazyVec ℕ ℓ (replicate C)
      -> Thunk 2 (LazyVec ℕ (ℓ + 1) (replicate (C + 2)))
    insert' n lNil         = return' $ lCons n $ return' lNil
    insert' n (lCons x xs) = step $ helper n x (n ≤?> x) xs

    helper (C, ℓ : ℕ)
      :  (n, x : ℕ)
      -> n ≤ x ∔ x + 1 ≤ n
      -> Thunk C (LazyVec ℕ ℓ (replicate C))
      -> Thunk 1 (LazyVec ℕ (ℓ + 2) (replicate (C + 2)))
    helper n x (left _)  xs = return' $ lCons n $
        slower plusIncreasingLeft $ slowerLazyVec {1} $ lCons x xs
    helper n x (right _) xs = step $ return $ lCons x (insert' n =<< xs)

insertionSort (ℓ : ℕ)
  : Vec ℓ ℕ -> Thunk (2 * ℓ) (LazyVec ℕ ℓ (replicate (2 * ℓ)))
insertionSort []        = return lNil
insertionSort (x :: xs) = insert x =<< insertionSort xs

headLazyVec (ℓ : ℕ) (Cs : Vec (ℓ + 1) ℕ) (a : Set)
  : LazyVec a (ℓ + 1) Cs -> a
headLazyVec (lCons x _) = x

minimum' (ℓ : ℕ) : Vec (ℓ + 1) ℕ -> Thunk (2 * (ℓ + 1)) ℕ
minimum' xs = return ∘ headLazyVec =<< insertionSort xs
