-- # import eq
-- # import prelude

sig Semigroup (a : Set) : Set where
  zero                : a
  (+)                 : a -> a -> a
  assoc (x, y, z : a) : x + (y + z) ≡ (x + y) + z

sig LeftMonoid (a : Set) : Set extends Semigroup where
  leftUnit (x : a) : zero + x ≡ x

sig RightMonoid (a : Set) : Set extends Semigroup where
  rightUnit (x : a) : x + zero ≡ x

sig Monoid (a : Set) : Set extends LeftMonoid, RightMonoid where

-- m₁ ⇰ m₂ is a homomorphism from m₁ to m₂.

sig (⇰) {a, b : Set} (m₁ : Semigroup a) (m₂ : Semigroup b) : Set where
  f    : a -> b
  zero : f m₁.zero ≡ m₂.zero
  plus : (x, y : a) -> f (m₁.(+) x y) ≡ m₂.(+) (f x) (f y)

-- Identity homomorphism.

reflHom (a : Set) (m : Semigroup a)
  : m ⇰ m
reflHom = struct
  f        = id
  zero     = refl
  plus _ _ = refl

-- Transitivity of homomorphisms.

transHom (a, b, c : Set)
         (m₁ : Semigroup a) (m₂ : Semigroup b) (m₃ : Semigroup c)
  : m₁ ⇰ m₂ -> m₂ ⇰ m₃ -> m₁ ⇰ m₃
transHom h₁ h₂ = struct
  f        = h₂.f ∘ h₁.f
  zero     = cong h₂.f h₁.zero
               `trans`
             h₂.zero
  plus x y = cong h₂.f (h₁.plus x y)
               `trans`
             h₂.plus (h₁.f x) (h₁.f y)
