------------------------------------------------------------------------
-- The types give correct amortised upper time bounds
------------------------------------------------------------------------

module Correct where

open import MainResults
open import Data.Nat
open import Data.Nat.Properties
open import Data.List renaming ([_] to List)
open import Data.Function
open import Relation.Binary hiding (_⇒_)
open import Relation.Binary.PropositionalEquality
open ≡-Reasoning
  using ()
  renaming (begin_ to ≡-begin_; _≡⟨_⟩_ to _≡-⟨_⟩_; _∎ to _≡-∎)
open ≤-Reasoning
open import TypeSystem
open import Erasure
open import Substitution
open VarSubst
open import Value
open import OperationalSemantics.Binding
open Heap↦
open import OperationalSemantics.GPay
open import TypeTriple
open import PayCtxt
open import OperationalSemantics.Thunked
open import Correct.Lemmas
import OperationalSemantics.Complete as C using (complete)

------------------------------------------------------------------------
-- Some lemmas

-- Credit is preserved by the application of substitutions.

credit↦-/↦-pres
  :  forall {Γ Δ τ} (b : Γ ↦ τ) (ρ : Γ ⇒ Δ)
  -> credit↦ (b /↦ ρ) ≡ credit↦ b
credit↦-/↦-pres (↦ⁿˢ ns e) ρ = ≡-refl

-- Some lemmas relating time with other operations.

time-Thunks-lemma :  forall ns {τ}
                  -> time (Thunks ns τ) ≡ sum ns + time τ
time-Thunks-lemma []               = ≡-byDef
time-Thunks-lemma (n ∷ ns) {τ = τ} = ≡-begin
  n + time (Thunks ns τ)
                          ≡-⟨ ≡-cong (_+_ n) (time-Thunks-lemma ns) ⟩
  n + (sum ns + time τ)
                          ≡-⟨ ≡-sym (lem₄ n (sum ns) (time τ)) ⟩
  sum (n ∷ ns) + time τ
                          ≡-∎

time-C-lemma
  :  forall C {m₁} m₂ {τ}
  -> time (C [ Thunk m₁ τ ]) ≤ time (C [ Thunk (m₂ + m₁) τ ])
time-C-lemma ● {m₁ = m₁}  m₂ = n≤m+n m₂ m₁ +-mono byDef
time-C-lemma (const● σ)   m₂ = byDef
time-C-lemma (C ●× σ)     m₂ = byDef
time-C-lemma (σ ×● C)     m₂ = byDef
time-C-lemma (Thunk● n C) m₂ = byDef {x = n} +-mono time-C-lemma C m₂

time-C-lemma₂
  :  forall C {m₁} m₂ {τ}
  -> time (C [ Thunk m₁ τ ]) ≤ m₂ + time (C [ Thunk (m₁ ∸ m₂) τ ])
time-C-lemma₂ ● {m₁ = m₁} m₂ {τ = τ} = begin
  m₁ + time τ
                           ≤⟨ n≤m+n∸m m₂ m₁ +-mono byDef ⟩
  m₂ + (m₁ ∸ m₂) + time τ
                           ≡⟨ lem₄ m₂ (m₁ ∸ m₂) _ ⟩
  m₂ + (m₁ ∸ m₂ + time τ)
                           ∎
time-C-lemma₂ (const● σ)             m₂         = n≤m+n m₂ _
time-C-lemma₂ (C ●× σ)               m₂         = z≤n
time-C-lemma₂ (σ ×● C)               m₂         = z≤n
time-C-lemma₂ (Thunk● n C) {m₁ = m₁} m₂ {τ = τ} = begin
  n + time (C [ Thunk m₁ τ ])
                                             ≤⟨ byDef {x = n}
                                                  +-mono
                                                time-C-lemma₂ C m₂ ⟩
  n + (m₂ + time (C [ Thunk (m₁ ∸ m₂) τ ]))
                                             ≡⟨ lem₁₄ n m₂ _ ⟩
  m₂ + (n + time (C [ Thunk (m₁ ∸ m₂) τ ]))
                                             ∎

------------------------------------------------------------------------
-- Correctness of OperationalSemantics.GPay

gpay↦-lemma
  :  forall {Γ σ n} C m (b : Γ ↦ C [ Thunk n σ ])
  -> credit↦ (gpay↦ C m b) ≡ m + credit↦ b
gpay↦-lemma C m (↦ⁿˢ ns e) = ≡-refl

mutual

  gpayResult∋-lemma
    :  forall {Γ} n₁ n₂ m₁ m₂ C (Σ₁ : Heap Γ) {σ τ}
              (x : Γ ∋ τ) (eq : τ ≡ C [ Thunk m₁ σ ])
    -> n₁ + credit Σ₁ ≤ n₂
    -> n₁ + credit (GPay.Σ₂ (gpayResult∋ C m₂ Σ₁ x eq)) ≤ m₂ + n₂
  gpayResult∋-lemma n₁ n₂ m₁ m₂ C Σ₁ {σ = σ}
                    (gwait∋ {n₁ = n'} C₁ n x) eq hyp =
    gpayResult∋-gwait-lemma ● m₁ m₂ n n' n₁ n₂ Σ₁ x
                            (complete (⟨_,_,_⟩ C₁ _ _)
                                      (⟨_,_,_⟩ C  _ _) eq)
                            hyp
  gpayResult∋-lemma n₁ n₂ m₁ m₂ C (Σ₁ ▷ʰ b) vz ≡-refl hyp = begin
    n₁ + (credit↦ (gpay↦ C m₂ b) + credit Σ₁)
                                               ≡⟨ ≡-byDef {x = n₁}
                                                    ⟨ ≡-cong₂ _+_ ⟩
                                                  (gpay↦-lemma C m₂ b
                                                     ⟨ ≡-cong₂ _+_ ⟩
                                                   ≡-byDef) ⟩
    n₁ + (m₂ + credit↦ b + credit Σ₁)
                                               ≡⟨ lem₂₁ n₁ m₂ (credit↦ b) _ ⟩
    m₂ + (n₁ + (credit↦ b + credit Σ₁))
                                               ≤⟨ byDef {x = m₂} +-mono hyp ⟩
    m₂ + n₂
                                               ∎
  gpayResult∋-lemma n₁ n₂ m₁ m₂ C (Σ₁ ▷ʰ b) (vs x) ≡-refl hyp = begin
    n₁ + (c + c')
                   ≡⟨ ≡-sym (lem₄ n₁ c _) ⟩
    n₁ + c + c'
                   ≤⟨ gpayResult∋-lemma (n₁ + c) n₂ m₁ m₂ C
                                        Σ₁ x ≡-refl hyp' ⟩
    m₂ + n₂
                   ∎
    where
    c    = credit↦ (b /↦ GPay.ρ (gpayResult∋ C m₂ Σ₁ x ≡-refl))
    c'   = credit (GPay.Σ₂ (gpayResult∋ C m₂ Σ₁ x ≡-refl))
    hyp' = begin
      n₁ + c + credit Σ₁            ≡⟨ (≡-byDef {x = n₁}
                                         ⟨ ≡-cong₂ _+_ ⟩
                                        credit↦-/↦-pres b _)
                                         ⟨ ≡-cong₂ _+_ ⟩
                                       ≡-byDef ⟩
      n₁ + credit↦ b + credit Σ₁    ≡⟨ lem₄ n₁ (credit↦ b) _ ⟩
      n₁ + (credit↦ b + credit Σ₁)  ≤⟨ hyp ⟩
      n₂                            ∎

  gpayResult∋-gwait-lemma
    :  forall C {C₁ C₂} (m₁ m₂ k₁ k₂ n₁ n₂ : ℕ) {Γ σ₁ σ₂} Σ₁
    -> (x : Γ ∋ C ⁻¹∘ C₁ [ Thunk k₂ σ₁ ])
    -> (eq : ⟨ C₁ , k₁ + k₂ , σ₁ ⟩ ≜ ⟨ C₂ , m₁ , σ₂ ⟩)
    -> n₁ + credit Σ₁ ≤ n₂
    -> n₁ + credit (GPay.Σ₂ (gpayResult∋-gwait C m₂ k₁ k₂ m₁ Σ₁ x eq)) ≤
       m₂ + n₂
  gpayResult∋-gwait-lemma C m₁ m₂ k₁ k₂ n₁ n₂ Σ₁ x (≜c- {C = C₂}) hyp =
    gpayResult∋-lemma n₁ n₂ m₁ m₂ (C ⁻¹∘ C₂) Σ₁ x
                      (⁻¹-const●-lemma C C₂) hyp
  gpayResult∋-gwait-lemma C m₁ m₂ k₁ k₂ n₁ n₂ Σ₁ x
                          (≜-c {C = C₁} {σ₁ = σ₁}) hyp =
    gpayResult∋-lemma n₁ n₂ m₁ m₂ (C ⁻¹∘ const● (C₁ [ Thunk k₂ σ₁ ]))
                      Σ₁ {σ = σ₁} x (≡-sym (⁻¹-const●-lemma C C₁)) hyp
  gpayResult∋-gwait-lemma C .(k₁ + k₂) m₂ k₁ k₂ n₁ n₂ Σ₁ x ≜● hyp =
    gpayResult∋-lemma n₁ n₂ k₂ m₂ (C ⁻¹) Σ₁ x ≡-refl hyp
  gpayResult∋-gwait-lemma C m₁ m₂ k₁ k₂ n₁ n₂ Σ₁ x (≜●T {C = C'}) hyp =
    gpayResult∋-lemma n₁ n₂ m₁ m₂ (C ⁻¹∘ Thunk● k₂ C') Σ₁ x
                      (⁻¹_[]-lemma C (Thunk● k₂ C')) hyp
  gpayResult∋-gwait-lemma C m₁ m₂ k₁ k₂ n₁ n₂ Σ₁ x (≜T● {C = C'}) hyp =
    gpayResult∋-lemma n₁ n₂ m₁ m₂ (C ⁻¹) Σ₁ x
                      (≡-sym (⁻¹_[]-lemma C (Thunk● m₁ C'))) hyp
  gpayResult∋-gwait-lemma C m₁ m₂ k₁ k₂ n₁ n₂ Σ₁ x (≜TT {n = n} eq) hyp =
    gpayResult∋-gwait-lemma (Thunk● n C) m₁ m₂ k₁ k₂ n₁ n₂ Σ₁ x eq hyp
  gpayResult∋-gwait-lemma C m₁ m₂ k₁ k₂ n₁ n₂ Σ₁ x
                          (≜●×² {σ = σ} eq) hyp =
    gpayResult∋-gwait-lemma (C ●× σ) m₁ m₂ k₁ k₂ n₁ n₂ Σ₁ x eq hyp
  gpayResult∋-gwait-lemma C m₁ m₂ k₁ k₂ n₁ n₂ Σ₁ x
                          (≜×●² {σ = σ} eq) hyp =
    gpayResult∋-gwait-lemma (σ ×● C) m₁ m₂ k₁ k₂ n₁ n₂ Σ₁ x eq hyp
  gpayResult∋-gwait-lemma C m₁ m₂ k₁ k₂ n₁ n₂ Σ₁ x
                          (≜●××● {C₁ = C₁} {C₂ = C₂} {σ₁ = σ₁}) hyp =
    gpayResult∋-lemma n₁ n₂ m₁ m₂ (C ⁻¹∘ (C₁ [ Thunk k₂ σ₁ ] ×● C₂))
                      Σ₁ x (≡-sym (×-lemma C C₁ C₂)) hyp
  gpayResult∋-gwait-lemma C m₁ m₂ k₁ k₂ n₁ n₂ Σ₁ x
                          (≜×●●× {C₁ = C₁} {C₂ = C₂} {σ₂ = σ₂}) hyp =
    gpayResult∋-lemma n₁ n₂ m₁ m₂ (C ⁻¹∘ (C₁ ●× C₂ [ Thunk k₂ σ₂ ]))
                      Σ₁ x (×-lemma C C₁ C₂) hyp

gpayResult-lemma
  :  forall {Γ} n₁ n₂ m₁ m₂ C (Σ₁ : Heap Γ) τ
            (v : Γ ⊢⇓ C [ Thunk m₁ τ ])
  -> n₁ + credit Σ₁ ≤ time (C [ Thunk m₁ τ ]) + n₂
  -> n₁ + credit (GPay.Σ₂ (gpayResult C m₂ Σ₁ v)) ≤
     m₂ + time (C [ Thunk (m₁ ∸ m₂) τ ]) + n₂
gpayResult-lemma n₁ n₂ m₁ m₂ ● Σ₁ τ (return⋆⇓ v) hyp = begin
  n₁ + credit Σ₁
                                            ≤⟨ hyp ⟩
  time (● [ Thunk m₁ τ ]) + n₂
                                            ≤⟨ time-C-lemma₂ ● {m₁ = m₁} m₂ {τ = τ}
                                                 +-mono
                                               byDef ⟩
  m₂ + time (● [ Thunk (m₁ ∸ m₂) τ ]) + n₂
                                            ∎
gpayResult-lemma n₁ n₂ m₁ m₂ (const● σ) Σ₁ τ v hyp = begin
  n₁ + credit Σ₁
                                            ≤⟨ hyp ⟩
  time (const● σ [ Thunk m₁ τ ]) + n₂
                                            ≤⟨ n≤m+n m₂ _ +-mono byDef ⟩
  m₂ + time (const● σ [ Thunk m₁ τ ]) + n₂
                                            ∎
gpayResult-lemma n₁ n₂ m₁ m₂ (C ●× σ₂) Σ₁ τ (x ,⇓ y) hyp = begin
  n₁ + credit (GPay.Σ₂ (gpayResult∋ C m₂ Σ₁ x ≡-refl))
               ≤⟨ gpayResult∋-lemma n₁ n₂ m₁ m₂ C Σ₁ x ≡-refl hyp ⟩
  m₂ + n₂
               ≡⟨ lem₂₀ m₂ n₂ ⟩
  m₂ + 0 + n₂
               ∎
gpayResult-lemma n₁ n₂ m₁ m₂ (σ₁ ×● C) Σ₁ τ (x ,⇓ y) hyp = begin
  n₁ + credit (GPay.Σ₂ (gpayResult∋ C m₂ Σ₁ y ≡-refl))
               ≤⟨ gpayResult∋-lemma n₁ n₂ m₁ m₂ C Σ₁ y ≡-refl hyp ⟩
  m₂ + n₂
               ≡⟨ lem₂₀ m₂ n₂ ⟩
  m₂ + 0 + n₂
               ∎
gpayResult-lemma n₁ n₂ m₁ m₂ (Thunk● k C) Σ₁ τ (return⋆⇓ v) hyp = begin
  n₁ + credit (GPay.Σ₂ (gpayResult C m₂ Σ₁ v))
                                                  ≤⟨ gpayResult-lemma
                                                       n₁ (k + n₂) m₁ m₂ C Σ₁ τ v hyp' ⟩
  m₂ + time (C [ Thunk (m₁ ∸ m₂) τ ]) + (k + n₂)
                                                  ≡⟨ lem₁₅ m₂ _ k _ ⟩
  m₂ + (k + time (C [ Thunk (m₁ ∸ m₂) τ ])) + n₂
                                                  ∎
  where
  hyp' = begin
    n₁ + credit Σ₁                     ≤⟨ hyp ⟩
    k + time (C [ Thunk m₁ τ ]) + n₂   ≡⟨ lem₁ k _ _ ⟩
    time (C [ Thunk m₁ τ ]) + (k + n₂) ∎

------------------------------------------------------------------------
-- The types are correct

-- The two gwait cases have the same form, as captured by this lemma.

correct-gwait-lemma
  :  forall n₁ n₂ n₃ n₄ C τ
  -> n₁ ≤ time (C [ Thunk n₂ τ ]) + n₃
  -> n₁ ≤ time (C [ Thunk (n₄ + n₂) τ ]) + n₃
correct-gwait-lemma n₁ n₂ n₃ n₄ C τ hyp = begin
  n₁
                                       ≤⟨ hyp ⟩
  time (C [ Thunk n₂ τ ]) + n₃
                                       ≤⟨ time-C-lemma C n₄ +-mono byDef ⟩
  time (C [ Thunk (n₄ + n₂) τ ]) + n₃
                                       ∎

-- The three last cases of the main proof all have the same form, as
-- captured by this lemma.

correct•-lemma :  forall n₁ n₂ n₃ n₄ n₅ n₆
               -> n₂ + n₃ ≤ n₆ + n₄
               -> n₁ + n₄ ≤ n₅
               -> 1 + (n₁ + n₂ + n₃) ≤ 1 + (n₆ + n₅)
correct•-lemma n₁ n₂ n₃ n₄ n₅ n₆ hyp₁ hyp₂ = begin
  1 + (n₁ + n₂ + n₃)
                        ≡⟨ lem₄ (1 + n₁) n₂ n₃ ⟩
  (1 + n₁) + (n₂ + n₃)
                        ≤⟨ byDef {x = 1 + n₁} +-mono hyp₁ ⟩
  (1 + n₁) + (n₆ + n₄)
                        ≡⟨ lem₁₉ 1 n₁ n₆ n₄ ⟩
  (1 + n₆) + (n₁ + n₄)
                        ≤⟨ byDef {x = 1 + n₆} +-mono hyp₂ ⟩
  (1 + n₆) + n₅
                        ≡⟨ ≡-byDef ⟩
  1 + (n₆ + n₅)
                        ∎

mutual

  -- The compositional correctness result.

  correct : Correctness

  -- Values.

  correct eval-λ       = byDef
  correct eval-λ²      = byDef
  correct eval-z       = byDef
  correct eval-s       = byDef
  correct eval-natrec  = byDef
  correct eval-,       = byDef
  correct eval-uncurry = byDef

  -- Variables.

  correct (eval-vz {τ = τ} {ms = ms} {n = n} {Σ₁ = Σ₁} {Σ₂ = Σ₂} d) =
                                    begin
    n + credit Σ₂
                                    ≤⟨ correct d ⟩
    time (Thunks ms τ) + credit Σ₁
                                    ≡⟨ time-Thunks-lemma ms
                                         ⟨ ≡-cong₂ _+_ ⟩
                                       ≡-byDef ⟩
    sum ms + time τ + credit Σ₁
                                    ≡⟨ lem₁ (sum ms) (time τ) (credit Σ₁) ⟩
    time τ + (sum ms + credit Σ₁)
                                    ∎

  correct (eval-vs {τ = τ} {n = n} {ρ = ρ}
                   {Σ₁ = Σ₁} {Σ₂ = Σ₂} {b = b} d) =
                                        begin
    n + (credit↦ (b /↦ ρ) + credit Σ₂)
                                        ≡⟨ lem₂ n (credit↦ (b /↦ ρ)) (credit Σ₂) ⟩
    n + credit Σ₂ + credit↦ (b /↦ ρ)
                                        ≡⟨ ≡-byDef {x = n + credit Σ₂}
                                             ⟨ ≡-cong₂ _+_ ⟩
                                           credit↦-/↦-pres b ρ ⟩
    n + credit Σ₂ + credit↦ b
                                        ≤⟨ correct d +-mono byDef ⟩
    time τ + credit Σ₁ + credit↦ b
                                        ≡⟨ lem₃ (time τ) (credit Σ₁) (credit↦ b) ⟩
    time τ + (credit↦ b + credit Σ₁)
                                        ∎

  correct (eval-gwait∋ {τ = τ} {m₁ = m₁} {m₂ = m₂} {n = n}
                       {Σ₁ = Σ₁} {Σ₂ = Σ₂} {C = C} {v = v} d) =
    correct-gwait-lemma (n + credit Σ₂) m₁ (credit Σ₁) m₂ C τ
                        (correct d)

  -- Primitives.

  correct (eval-return {τ = τ} {n = n} {Σ₁ = Σ₁} {Σ₂ = Σ₂} d) =
                        begin
    n + credit Σ₂
                        ≤⟨ correct d ⟩
    time τ + credit Σ₁
                        ∎

  correct (eval-√ {τ = τ} {m = m} {n = n} {Σ₁ = Σ₁} {Σ₂ = Σ₂} d) =
                                  begin
    n + credit Σ₂
                                  ≤⟨ correct d ⟩
    m + time τ + credit Σ₁
                                  ≤⟨ n≤1+n _ ⟩
    suc (m + time τ + credit Σ₁)
                                  ∎

  correct (eval-pay {τ = τ} {m₁ = m₁} {m₂ = m₂} {n = n}
                    {Σ₁ = Σ₁} {Σ₂ = Σ₂} {v = v} d) =
                                                     begin
    n + credit Σ₂
                                                     ≤⟨ correct d ⟩
    m₁ + time τ + credit Σ₁
                                                     ≡⟨ lem₄ m₁ _ _ ⟩
    m₁ + (time τ + credit Σ₁)
                                                     ≤⟨ n≤m+n∸m m₂ m₁ +-mono byDef ⟩
    m₂ + (m₁ ∸ m₂) + (time τ + credit Σ₁)
                                                     ≡⟨ lem₆ m₂ _ (time τ) _ ⟩
    m₂ + (m₁ ∸ m₂ + time τ) + credit Σ₁
                                                     ≡⟨ ≡-byDef ⟩
    time (Thunk m₂ (Thunk (m₁ ∸ m₂) τ)) + credit Σ₁
                                                     ∎

  correct (eval-gpay {τ = τ} {m₁ = m₁} {m₂ = m₂} {n = n}
                     {Σ₁ = Σ₁} {Σ₂ = Σ₂} {C = C} {v = v} d) =
    gpayResult-lemma n (credit Σ₁) m₁ m₂ C Σ₂ τ v (correct d)

  correct (eval->>= {τ = τ} {m₁ = m₁} {m₂ = m₂} {n₁ = n₁} {n₂ = n₂}
                    {Σ₁ = Σ₁} {Σ₂ = Σ₂} {Σ₃ = Σ₃} d₁ d₂) =
                                               begin
    n₁ + n₂ + credit Σ₃
                                               ≡⟨ lem₄ n₁ _ _ ⟩
    n₁ + (n₂ + credit Σ₃)
                                               ≤⟨ byDef {x = n₁} +-mono correct• d₂ ⟩
    n₁ + (m₂ + time τ + (m₁ + 0 + credit Σ₂))
                                               ≡⟨ lem₁₇ n₁ m₂ _ m₁ _ ⟩
    m₁ + m₂ + time τ + (n₁ + credit Σ₂)
                                               ≤⟨ byDef {x = m₁ + m₂ + time τ}
                                                    +-mono
                                                  correct d₁ ⟩
    m₁ + m₂ + time τ + credit Σ₁
                                               ∎

  correct (eval-gwait {τ = τ} {m₁ = m₁} {m₂ = m₂} {n = n}
                      {Σ₁ = Σ₁} {Σ₂ = Σ₂} {C = C} {v = v} d) =
    correct-gwait-lemma (n + credit Σ₂) m₁ (credit Σ₁) m₂ C τ
                        (correct d)

  correct (eval-fix {σ = σ} {m = m} {n = n} {Σ₁ = Σ₁} {Σ₂ = Σ₂} d) =
                                  begin
    (1 + n) + credit Σ₂
                                  ≡⟨ ≡-byDef ⟩
    1 + (n + credit Σ₂)
                                  ≤⟨ byDef {x = 1} +-mono correct d ⟩
    1 + (m + time σ + credit Σ₁)
                                  ≡⟨ ≡-byDef ⟩
    (1 + m) + time σ + credit Σ₁
                                  ∎

  -- Applications.

  correct (eval-· {τ = τ} {n₁ = n₁} {n₂ = n₂}
                  {Σ₁ = Σ₁} {Σ₂ = Σ₂} {Σ₃ = Σ₃} d₁ d₂) =
                               begin
    n₁ + n₂ + credit Σ₃
                               ≡⟨ lem₄ n₁ _ _ ⟩
    n₁ + (n₂ + credit Σ₃)
                               ≤⟨ byDef {x = n₁} +-mono correct• d₂ ⟩
    n₁ + (time τ + credit Σ₂)
                               ≡⟨ lem₁₄ n₁ (time τ) _ ⟩
    time τ + (n₁ + credit Σ₂)
                               ≤⟨ byDef {x = time τ} +-mono correct d₁ ⟩
    time τ + credit Σ₁
                               ∎

  correct•
    :  forall {Γ Δ σ τ Σ₁ Σ₂ n} {ρ : _ ⇒ Δ} {e₁ : Γ ⊢⇓ σ → τ} {v}
    -> Σ₁ ∣ e₁ •⇊ ρ ⋉ Σ₂ ∣ v √ n
    -> n + credit Σ₂ ≤ time τ + credit Σ₁

  -- Successor.

  correct• eval-·-s = byDef

  -- Lambda.

  correct• (eval-·-λ {m = m} {n = n} {τ = τ}
                     {Σ₁ = Σ₁} {Σ₂ = Σ₂} d) =
                                  begin
    1 + n + credit Σ₂
                                  ≡⟨ lem₄ 1 n _ ⟩
    1 + (n + credit Σ₂)
                                  ≤⟨ byDef {x = 1} +-mono correct d ⟩
    1 + (m + time τ + credit Σ₁)
                                  ≡⟨ lem₁₈ 1 m _ _ ⟩
    1 + m + time τ + credit Σ₁
                                  ∎

  correct• eval-·-λ² = byDef

  -- Natrec.

  correct• (eval-·-natrec-zero
              {m = m} {n₁ = n₁} {n₂ = n₂} {τ = τ}
              {Σ₁ = Σ₁} {Σ₂ = Σ₂} {Σ₃ = Σ₃} d₁ d₂) =
    correct•-lemma
      n₁ n₂ (credit Σ₃) (credit Σ₂) (credit Σ₁) (m + time τ)
      (correct d₂) (correct d₁)

  correct• (eval-·-natrec-suc
              {m = m} {n₁ = n₁} {n₂ = n₂} {τ = τ}
              {Σ₁ = Σ₁} {Σ₂ = Σ₂} {Σ₃ = Σ₃} d₁ d₂) =
    correct•-lemma
      n₁ n₂ (credit Σ₃) (credit Σ₂) (credit Σ₁) (m + time τ)
      (correct d₂) (correct d₁)

  -- Uncurry.

  correct• (eval-·-uncurry
              {m = m} {n₁ = n₁} {n₂ = n₂} {τ = τ}
              {Σ₁ = Σ₁} {Σ₂ = Σ₂} {Σ₃ = Σ₃} d₁ d₂) =
    correct•-lemma
      n₁ n₂ (credit Σ₃) (credit Σ₂) (credit Σ₁) (m + time τ)
      (correct d₂) (correct d₁)

-- The simplified correctness result.

correct' : Correctness'
correct' {τ = τ} Σ₁ {n = n} e d with C.complete _ Σ₁ e d
... | ∃thunked _ ≡-refl _ ≡-refl Σ₂ ≡-refl _ ≡-refl d' = begin
  n                   ≤⟨ n≤m+n (credit Σ₂) n ⟩
  credit Σ₂ + n       ≡⟨ lem₂₂ (credit Σ₂) n ⟩
  n + credit Σ₂       ≤⟨ correct d' ⟩
  time τ + credit Σ₁  ∎
