------------------------------------------------------------------------
-- Counterexample showing that a certain simplification of the
-- correctness result is not possible
------------------------------------------------------------------------

-- This module provides an example showing that
--
--   Γ₁ ⊢ e : τ  ∧  Γ₁ ⊢ Σ₁  ∧
--   Γ₂ ⊢ v : τ  ∧  Γ₂ ⊢ Σ₂  ∧
--   ⌜ Σ₁ ⌝ | ⌜ e ⌝ ⇓ⁿ ⌜ Σ₂ ⌝ | ⌜ v ⌝
--
--   ⇒ credit Σ₂ + n ≤ credit Σ₁ + time τ
--
-- is not true. This suggests that it is not possible to express the
-- correctness result, in full generality, using only the simple
-- semantics. As shown in Correct the following simplification is
-- possible, though:
--
--   Γ₁ ⊢ e : τ  ∧  ⌜ Σ₁ ⌝ | ⌜ e ⌝ ⇓ⁿ Σ₂ | v  ⇒  n ≤ credit Σ₁ + time τ

module CounterExample where

open import Kind
open import TypeSystem
open import Erasure
open import Substitution
open VarSubst
open import Value
open import OperationalSemantics.Binding
open Heap↦
open Heap↦Erasure
open import OperationalSemantics.Simple
open import MainResults

open import Data.Nat
open import Relation.Nullary

-- Γ₁ = Γ₂.

Γ : Ctxt thunked
Γ = ε ▻ Nat

Σ : ℕ -> Heap Γ
Σ n = ∅ʰ ▷ʰ ↦ⁿ return⋆ {n = n} z

Σ₁ = Σ zero
Σ₂ = Σ (suc zero)

τ : Ty thunked
τ = Nat

e : Γ ⊢ τ
e = z

v : Γ ⊢⇓ τ
v = z⇓

n : ℕ
n = zero

d : ⌜ Σ₁ ⌝ʰ ∣ ⌜ e ⌝ ⇓ idˢ ⋉ ⌜ Σ₂ ⌝ʰ ∣ ⌜ v ⌝⇓ √ n
d = eval-z

counterexample : ¬ (credit Σ₂ + n ≤ credit Σ₁ + time τ)
counterexample ()
