------------------------------------------------------------------------
-- Kinds (simple or thunked)
------------------------------------------------------------------------

module Kind where

open import Data.Unit using (⊤; tt)
open import Data.Nat using (ℕ; suc)

-- The type of calculus: thunked or unthunked (simple).

data Kind : Set where
  simple  : Kind
  thunked : Kind

-- In the thunked type system natural numbers are used to keep track
-- of evaluation steps, but not in the simple system. By using Maybe ℕ
-- the two type systems can be merged in a nice way (see
-- TypeSystem.Thunk? and TypeSystem._⊢_). The η equality of the unit
-- type (⊤) is convenient here, since it ensures that the single unit
-- value does not need to be written out when it is hidden, even if
-- nothing depends on it.

Maybe : Set -> Kind -> Set
Maybe a simple  = ⊤
Maybe a thunked = a

-- A lifted version of suc.

suc' : forall k -> Maybe ℕ k -> Maybe ℕ k
suc' simple  _ = tt
suc' thunked n = suc n
