------------------------------------------------------------------------
-- Some boring cast lemmas
------------------------------------------------------------------------

module OperationalSemantics.Casts where

open import Kind
open import TypeSystem
open import OperationalSemantics.Binding
open Heap↦
open import OperationalSemantics.Simple
open import Substitution
open VarSubst
open import Value
open import Relation.Binary.PropositionalEquality
open import Relation.Binary.HeterogeneousEquality

------------------------------------------------------------------------
-- Casts

record AllEqual
  {Γ¹ Δ¹ : Ctxt simple} {τ¹ : _}
  (Σ₁¹ : Heap Γ¹) (ρ¹ : Γ¹ ⇒ Δ¹) (Σ₂¹ : Heap Δ¹)
  (t¹ : Γ¹ ⊢ τ¹) (v¹ : Δ¹ ⊢⇓ τ¹)
  {Γ² Δ² : Ctxt simple} {τ² : _}
  (Σ₁² : Heap Γ²) (ρ² : Γ² ⇒ Δ²) (Σ₂² : Heap Δ²)
  (t² : Γ² ⊢ τ²) (v² : Δ² ⊢⇓ τ²) : Set where
  field
    Γ-eq  : Γ¹  ≡ Γ²
    τ-eq  : τ¹  ≡ τ²
    Σ₁-eq : Σ₁¹ ≅ Σ₁²
    t-eq  : t¹  ≅ t²
    Δ-eq  : Δ¹  ≡ Δ²
    ρ-eq  : ρ¹  ≅ ρ²
    Σ₂-eq : Σ₂¹ ≅ Σ₂²
    v-eq  : v¹  ≅ v²

open AllEqual

abstract

  cast
    :  forall {Γ Γ' Δ Δ' τ τ' Σ₁ Σ₁' ρ ρ' Σ₂ Σ₂' n}
              {t  : Γ  ⊢ τ}  {v  : Δ  ⊢⇓ τ}
              {t' : Γ' ⊢ τ'} {v' : Δ' ⊢⇓ τ'}
    -> Σ₁  ∣ t  ⇓ ρ  ⋉ Σ₂  ∣ v  √ n
    -> AllEqual Σ₁ ρ Σ₂ t v Σ₁' ρ' Σ₂' t' v'
    -> Σ₁' ∣ t' ⇓ ρ' ⋉ Σ₂' ∣ v' √ n
  cast d eq with Γ-eq eq | τ-eq eq | Σ₁-eq eq | t-eq eq
               | Δ-eq eq | ρ-eq eq | Σ₂-eq eq | v-eq eq
  ... | ≡-refl | ≡-refl | ≅-refl | ≅-refl
      | ≡-refl | ≅-refl | ≅-refl | ≅-refl = d

  cast•
    :  forall {Γ Δ σ τ Σ₁ Σ₁' ρ ρ' Σ₂ n}
              {v₀ : Γ ⊢⇓ σ → τ} {v : Δ ⊢⇓ τ}
    -> Σ₁  ∣ v₀ •⇓ ρ  ⋉ Σ₂ ∣ v √ n
    -> Σ₁ ≡ Σ₁' -> ρ ≡ ρ'
    -> Σ₁' ∣ v₀ •⇓ ρ' ⋉ Σ₂ ∣ v √ n
  cast• d ≡-refl ≡-refl = d

------------------------------------------------------------------------
-- Partially defined AllEqual records

typesEqual
  :  forall {Γ Δ τ¹ τ² Σ₁} {ρ : Γ ⇒ Δ} {Σ₂ t¹ t²}
            {v¹ : Δ ⊢⇓ τ¹} {v² : Δ ⊢⇓ τ²}
  -> τ¹ ≡ τ² -> t¹ ≅ t² -> v¹ ≅ v²
  -> AllEqual Σ₁ ρ Σ₂ t¹ v¹ Σ₁ ρ Σ₂ t² v²
typesEqual ≡-refl ≅-refl ≅-refl = record
  { Γ-eq  = ≡-refl
  ; τ-eq  = ≡-refl
  ; Σ₁-eq = ≅-refl
  ; t-eq  = ≅-refl
  ; Δ-eq  = ≡-refl
  ; ρ-eq  = ≅-refl
  ; Σ₂-eq = ≅-refl
  ; v-eq  = ≅-refl
  }

inputsEqual
  :  forall {Γ Δ τ Σ₁¹ Σ₁²} {ρ : Γ ⇒ Δ} {Σ₂ t¹ t²} {v : Δ ⊢⇓ τ}
  -> Σ₁¹ ≡ Σ₁² -> t¹ ≡ t² -> AllEqual Σ₁¹ ρ Σ₂ t¹ v Σ₁² ρ Σ₂ t² v
inputsEqual ≡-refl ≡-refl = record
  { Γ-eq  = ≡-refl
  ; τ-eq  = ≡-refl
  ; Σ₁-eq = ≅-refl
  ; t-eq  = ≅-refl
  ; Δ-eq  = ≡-refl
  ; ρ-eq  = ≅-refl
  ; Σ₂-eq = ≅-refl
  ; v-eq  = ≅-refl
  }

termsEqual
  :  forall {Γ Δ τ Σ₁} {ρ : Γ ⇒ Δ} {Σ₂ t¹ t²} {v : Δ ⊢⇓ τ}
  -> t¹ ≡ t² -> AllEqual Σ₁ ρ Σ₂ t¹ v Σ₁ ρ Σ₂ t² v
termsEqual ≡-refl = inputsEqual ≡-refl ≡-refl

resultsEqual
  :  forall {Γ Δ τ Σ₁} {ρ¹ ρ² : Γ ⇒ Δ} {Σ₂¹ Σ₂² t} {v¹ v² : Δ ⊢⇓ τ}
  -> ρ¹ ≡ ρ² -> Σ₂¹ ≡ Σ₂² -> v¹ ≡ v²
  -> AllEqual Σ₁ ρ¹ Σ₂¹ t v¹ Σ₁ ρ² Σ₂² t v²
resultsEqual ≡-refl ≡-refl ≡-refl = record
  { Γ-eq  = ≡-refl
  ; τ-eq  = ≡-refl
  ; Σ₁-eq = ≅-refl
  ; t-eq  = ≅-refl
  ; Δ-eq  = ≡-refl
  ; ρ-eq  = ≅-refl
  ; Σ₂-eq = ≅-refl
  ; v-eq  = ≅-refl
  }

substsEqual
  :  forall {Γ Δ τ Σ₁} {ρ¹ ρ² : Γ ⇒ Δ} {Σ₂ t} {v : Δ ⊢⇓ τ}
  -> ρ¹ ≡ ρ² -> AllEqual Σ₁ ρ¹ Σ₂ t v Σ₁ ρ² Σ₂ t v
substsEqual ≡-refl = resultsEqual ≡-refl ≡-refl ≡-refl

------------------------------------------------------------------------
-- Lemmas about casts

abstract

  castEq
    :  forall {a : Set} {Γ Γ' Δ Δ' τ τ' Σ₁ Σ₁' ρ ρ' Σ₂ Σ₂' n}
              {t  : Γ  ⊢ τ}  {v  : Δ  ⊢⇓ τ}
              {t' : Γ' ⊢ τ'} {v' : Δ' ⊢⇓ τ'}
    -> (f :  forall {Γ Δ τ} {Σ₁} {t : Γ ⊢ τ}
                    {ρ : Γ ⇒ Δ} {Σ₂ v n}
          -> Σ₁ ∣ t ⇓ ρ ⋉ Σ₂ ∣ v √ n -> a)
    -> (d : Σ₁ ∣ t ⇓ ρ ⋉ Σ₂ ∣ v √ n)
    -> (eq : AllEqual Σ₁ ρ Σ₂ t v Σ₁' ρ' Σ₂' t' v')
    -> {x : a} -> f d ≡ x
    -> f (cast d eq) ≡ x
  castEq f d eq ≡-refl with Γ-eq eq | τ-eq eq | Σ₁-eq eq | t-eq eq
                          | Δ-eq eq | ρ-eq eq | Σ₂-eq eq | v-eq eq
  ... | ≡-refl | ≡-refl | ≅-refl | ≅-refl
      | ≡-refl | ≅-refl | ≅-refl | ≅-refl = ≡-refl

  cast•Eq
    :  forall {a : Set} {Γ Δ σ τ Σ₁ Σ₁' ρ ρ' Σ₂ n}
              {v₀ : Γ ⊢⇓ σ → τ} {v : Δ ⊢⇓ τ}
    -> (f :  forall {Γ Δ σ τ} {Σ₁} {v' : Γ ⊢⇓ σ → τ}
                    {ρ : Γ ▻ σ ⇒ Δ} {Σ₂ v n}
          -> Σ₁ ∣ v' •⇓ ρ ⋉ Σ₂ ∣ v √ n -> a)
    -> (d : Σ₁ ∣ v₀ •⇓ ρ ⋉ Σ₂ ∣ v √ n)
    -> (Σ₁-eq : Σ₁ ≡ Σ₁') -> (ρ-eq : ρ ≡ ρ')
    -> {x : a} -> f d ≡ x -> f (cast• d Σ₁-eq ρ-eq) ≡ x
  cast•Eq f d ≡-refl ≡-refl ≡-refl = ≡-refl
