------------------------------------------------------------------------
-- Completeness proof
------------------------------------------------------------------------

-- The thunked operational semantics is complete with respect to the
-- simple one.

module OperationalSemantics.Complete where

open import MainResults
open import Data.Nat
open import Kind
open import TypeSystem
open import Erasure
open import OperationalSemantics.Binding
open Heap↦
open Heap↦Erasure
open import Value
open import Substitution
open import Substitution.Lemmas
open VarSubst
open import OperationalSemantics.Simple  as S
open import OperationalSemantics.Thunked as T
open import OperationalSemantics.GPay
open import OperationalSemantics.GPay.Lemmas
open import OperationalSemantics.Casts
open import OperationalSemantics.Depth
open import Erasure.Lemmas
open import Erasure.Solver
open import Relation.Binary.PropositionalEquality
open ≡-Reasoning
open import Relation.Binary.HeterogeneousEquality
open import Data.Function

-- First a variant of ∃Thunked, used for applications.

data ∃Thunked• {Γ : Ctxt thunked} {Δ : Ctxt simple} {σ τ : Ty thunked}
       (Σ₁ : Heap (Γ ▻ σ)) (v₀ : Γ ⊢⇓ σ → τ)
       (ρ : ⌜ Γ ▻ σ ⌝▻ ⇒ Δ) (Σ₂ : Heap Δ) (v : Δ ⊢⇓ ⌜ τ ⌝⋆)
       (n : ℕ) : Set where
  ∃thunked• :  (Δ' : Ctxt thunked)
            -> (Δ-eq : Δ ≡ ⌜ Δ' ⌝▻)
            -> (ρ' : Γ ▻ σ ⇒ Δ')
            -> ≡-subst (_⇒_ _) Δ-eq ρ ≡ ⌜ ρ' ⌝⇒
            -> (Σ₂' : Heap Δ')
            -> ≡-subst Heap Δ-eq Σ₂ ≡ ⌜ Σ₂' ⌝ʰ
            -> (v' : Δ' ⊢⇓ τ)
            -> ≡-subst (\Δ -> Δ ⊢⇓ ⌜ τ ⌝⋆) Δ-eq v ≡ ⌜ v' ⌝⇓
            -> Σ₁ ∣ v₀ •⇊ ρ' ⋉ Σ₂' ∣ v' √ n
            -> ∃Thunked• Σ₁ v₀ ρ Σ₂ v n

-- The proof is by lexicographic induction over the depth of the
-- derivations and the structure of the "input" terms.

mutual

  complete : Completeness
  complete Γ Σ₁ e d = complete' Γ Σ₁ e d (depth d) ≡-refl

  complete∋ :  forall Γ {Δ τ} Σ₁ {ρ Σ₂ n} (x : Γ ∋ τ) {v : Δ ⊢⇓ ⌜ τ ⌝⋆}
            -> (deriv : ⌜ Σ₁ ⌝ʰ ∣ ⌜ var x ⌝ ⇓ ρ ⋉ Σ₂ ∣ v √ n)
            -> (d : Depth) -> depth deriv ≡ d
            -> ∃Thunked Σ₁ (var x) ρ Σ₂ v n

  complete∋ (Γ ▻ σ)  (Σ₁ ▷ʰ ↦ⁿˢ ms e₁) vz (S.eval-vz d) [0]         ()
  complete∋ (Γ ▻ σ)  (Σ₁ ▷ʰ ↦ⁿˢ ms e₁) vz (S.eval-vz d) (_ [max] _) ()
  complete∋ (Γ ▻ ._) (Σ₁ ▷ʰ ↦ⁿˢ {τ = τ} ms e₁) vz
            (S.eval-vz {v = v} d) ([suc] de) eq
    with complete' Γ Σ₁ e₁ {v = ≡-subst (_⊢⇓_ _) (≡-sym lemma₁) v}
                   (cast d allEq) de
                   (castEq depth d allEq (drop-[suc] eq))
    where
    lemma₁ = Thunks-lemma ms {τ = τ}
    lemma₂ = ≡-subst-removable (_⊢_ _) lemma₁ ⌜ e₁ ⌝
    lemma₃ = ≅-sym (≡-subst-removable (_⊢⇓_ _) (≡-sym lemma₁) v)
    allEq  = typesEqual (≡-sym lemma₁) lemma₂ lemma₃
  ... | ∃thunked Δ ≡-refl ρ' ≡-refl Σ' ≡-refl v' v-eq d' =
    ∃thunked (Δ ▻ τ) ≡-refl
             (ρ' ↑) lemma₁'
             (Σ' ▷ʰ ↦ valToTm v″) lemma₂'
             (v″ /⊢⇓ wk) lemma₃'
             (T.eval-vz d')
      where
      v″ = drop-return⋆⇓ ms v'

      lemma₁ = Thunks-lemma ms {τ = τ}

      lemma₁' = prove
        (|⌜ lift⇒ ρ' ⌝⇒| |↑|)
        (|⌜ _|↑| {σ = τ} (lift⇒ ρ') ⌝⇒|)
        ≡-refl

      lemma₄ = begin
        v
                                               ≡⟨ ≅-to-≡ (≅-sym (≡-subst-removable (_⊢⇓_ _) (≡-sym lemma₁) v)
                                                            ⟨ ≅-trans ⟩
                                                          ≅-sym (≡-subst-removable (_⊢⇓_ _) lemma₁ _)) ⟩
        ≡-subst (_⊢⇓_ _) lemma₁
          (≡-subst (_⊢⇓_ _) (≡-sym lemma₁) v)
                                               ≡⟨ ≡-cong (≡-subst (_⊢⇓_ _) lemma₁) v-eq ⟩
        ≡-subst (_⊢⇓_ _) lemma₁ ⌜ v' ⌝⇓
                                               ≡⟨ ⌜drop-return⋆⇓⌝-lemma ms v' ⟩
        ⌜ v″ ⌝⇓
                                               ∎

      lemma₂' = begin
        ⌜ Σ' ⌝ʰ ▷ʰ ↦ valToTm v
                                      ≡⟨ ≡-cong (\v -> ⌜ Σ' ⌝ʰ ▷ʰ ↦ (valToTm v)) lemma₄ ⟩
        ⌜ Σ' ⌝ʰ ▷ʰ ↦ valToTm ⌜ v″ ⌝⇓
                                      ≡⟨ prove (|⌜ liftʰ Σ' ⌝ʰ| |▷ʰ| |↦| |valToTm| |⌜ lift v″ ⌝⇓|)
                                               (|⌜ liftʰ Σ' |▷ʰ| |↦| |valToTm| (lift v″) ⌝ʰ|)
                                               ≡-refl ⟩
        ⌜ Σ' ▷ʰ ↦ valToTm v″ ⌝ʰ
                                      ∎

      lemma₃' = begin
        v /⊢⇓ wk
                                ≡⟨ ≡-cong₂ _/⊢⇓_ lemma₄ byDef ⟩
        ⌜ v″ ⌝⇓ /⊢⇓ wk
                                ≡⟨ prove (|⌜ lift v″ ⌝⇓| |/⊢⇓| |wk|)
                                         (|⌜ lift v″ |/⊢⇓| |wk| {σ = τ} ⌝⇓|)
                                         ≡-refl ⟩
        ⌜ v″ /⊢⇓ wk {σ = τ} ⌝⇓
                                ∎

  complete∋ (Γ ▻ σ) (Σ₁ ▷ʰ b) (vs x) (S.eval-vs d) [0]         ()
  complete∋ (Γ ▻ σ) (Σ₁ ▷ʰ b) (vs x) (S.eval-vs d) (_ [max] _) ()
  complete∋ (Γ ▻ σ) (Σ₁ ▷ʰ b) (vs x) (S.eval-vs d) ([suc] de)  eq
    with complete∋ Γ Σ₁ x d de (drop-[suc] eq)
  ... | ∃thunked Δ ≡-refl ρ' ≡-refl Σ' ≡-refl v' ≡-refl d' =
    ∃thunked (Δ ▻ σ) ≡-refl
             (ρ' ↑) lemma₁
             (Σ' ▷ʰ b /↦ ρ') lemma₂
             (v' /⊢⇓ wk) lemma₃
             (T.eval-vs d')
      where
      lemma₁ = prove
        (|⌜ lift⇒ ρ' ⌝⇒| |↑|)
        (|⌜ _|↑| {σ = σ} (lift⇒ ρ') ⌝⇒|)
        ≡-refl

      lemma₂ =
        prove (|⌜ liftʰ Σ' ⌝ʰ| |▷ʰ| |⌜ lift b ⌝↦| |/↦| |⌜ lift⇒ ρ' ⌝⇒|)
              (|⌜ liftʰ Σ' |▷ʰ| lift b |/↦| lift⇒ ρ' ⌝ʰ|)
              ≡-refl

      lemma₃ =
        prove (|⌜ lift v' ⌝⇓| |/⊢⇓| |wk|)
              (|⌜ lift v' |/⊢⇓| |wk| {σ = σ} ⌝⇓|)
              ≡-refl

  complete∋ Γ Σ₁ (gwait∋ {σ = τ} {n₁ = m₁} C m₂ x) {v = v} d de eq
    with complete∋ Γ Σ₁ x {v = v'} (cast d allEq) de
                   (castEq depth d allEq eq)
    where
    lem = erase-lemma {m = m₁} {n = m₂ + m₁} {τ = τ} C
    v'  = ≡-subst (_⊢⇓_ _) (≡-sym lem) v

    helper :  forall {σ} (eq : _ ≡ σ)
           -> var (≡-subst (_∋_ ⌜ Γ ⌝▻) eq ⌜ x ⌝∋) ≅ var ⌜ x ⌝∋
    helper ≡-refl = ≅-refl

    allEq : AllEqual
      ⌜ Σ₁ ⌝ʰ _ _ (var (≡-subst (_∋_ ⌜ Γ ⌝▻) lem ⌜ x ⌝∋)) v
      ⌜ Σ₁ ⌝ʰ _ _ ⌜ var x ⌝ v'
    allEq = record
      { Γ-eq  = ≡-refl
      ; τ-eq  = ≡-sym lem
      ; Σ₁-eq = ≅-refl
      ; t-eq  = helper lem
      ; Δ-eq  = ≡-refl
      ; ρ-eq  = ≅-refl
      ; Σ₂-eq = ≅-refl
      ; v-eq  = ≅-sym (≡-subst-removable (_⊢⇓_ _) (≡-sym lem) v)
      }
  ... | ∃thunked Δ ≡-refl ρ' ≡-refl Σ' ≡-refl v' v-eq d' =
    ∃thunked Δ ≡-refl ρ' ≡-refl Σ' ≡-refl
             (gwait⇓ C m₂ v') lemma
             (T.eval-gwait∋ {m₂ = m₂} {C = C} d')
    where
    lem'  = erase-lemma {m = m₁} {n = m₂ + m₁} {τ = τ} C
    cast₁ = ≡-subst (_⊢⇓_ ⌜ Δ ⌝▻) lem'
    cast₂ = ≡-subst (_⊢⇓_ ⌜ Δ ⌝▻) (≡-sym lem')
    lemma = begin
      v                    ≡⟨ ≅-to-≡ (≅-sym (≡-subst-removable (_⊢⇓_ _) (≡-sym lem') v)
                                        ⟨ ≅-trans ⟩
                                      ≅-sym (≡-subst-removable (_⊢⇓_ _) lem' _)) ⟩
      cast₁ (cast₂ v)      ≡⟨ ≡-cong cast₁ v-eq ⟩
      cast₁ ⌜ v' ⌝⇓        ≡⟨ prove (|≡-subst⇓| lem' |⌜ lift v' ⌝⇓|)
                                    |⌜ |gwait⇓| C m₂ (lift v') ⌝⇓|
                                    ≡-refl ⟩
      ⌜ gwait⇓ C m₂ v' ⌝⇓  ∎

  complete' :  forall Γ {Δ τ} Σ₁ {ρ Σ₂ n} (e : Γ ⊢ τ) {v : Δ ⊢⇓ ⌜ τ ⌝⋆}
            -> (deriv : ⌜ Σ₁ ⌝ʰ ∣ ⌜ e ⌝ ⇓ ρ ⋉ Σ₂ ∣ v √ n)
            -> (d : Depth) -> depth deriv ≡ d
            -> ∃Thunked Σ₁ e ρ Σ₂ v n

  -- Values.

  complete' Γ Σ₁ (λ e) S.eval-λ _ _ =
    ∃thunked Γ ≡-refl idˢ (⌜id⌝-lemma Γ) Σ₁ ≡-refl (λ⇓ e) ≡-refl
             T.eval-λ
  complete' Γ Σ₁ (λ² e) S.eval-λ² _ _ =
    ∃thunked Γ ≡-refl idˢ (⌜id⌝-lemma Γ) Σ₁ ≡-refl (λ²⇓ e) ≡-refl
             T.eval-λ²
  complete' Γ Σ₁ z S.eval-z _ _ =
    ∃thunked Γ ≡-refl idˢ (⌜id⌝-lemma Γ) Σ₁ ≡-refl z⇓ ≡-refl T.eval-z
  complete' Γ Σ₁ s S.eval-s _ _ =
    ∃thunked Γ ≡-refl idˢ (⌜id⌝-lemma Γ) Σ₁ ≡-refl s⇓ ≡-refl T.eval-s

  complete' Γ Σ₁ (natrec {n = n} {σ = σ} e₁ e₂) S.eval-natrec _ _ =
    ∃thunked (Γ ▻ Thunk n σ) ≡-refl wk (⌜wk⌝-lemma Γ (Thunk n σ))
             (Σ₁ ▷ʰ ↦ e₁) ≡-refl (natrec⇓ vz (e₂ /⊢ wk ↑ ↑))
             lemma T.eval-natrec
    where
    lemma = prove
      (|natrec⇓| (lift vz) (|⌜ lift e₂ ⌝| |/⊢| |wk| |↑| |↑|))
      (|natrec⇓| (lift vz)
                 |⌜ lift e₂ |/⊢|
                    _|↑| {σ = Thunk (suc n) σ}
                         (_|↑| {σ = Nat}
                               (|wk| {Γ = Γ} {σ = Thunk n σ})) ⌝|)
      ≡-refl

  complete' Γ Σ₁ (_,_ {σ = σ₁} {τ = σ₂} e₁ e₂) S.eval-, _ _ =
    ∃thunked (Γ ▻ σ₁ ▻ σ₂) ≡-refl (wk ∘ˢ wk) lemma₁
             (Σ₁ ▷ʰ ↦ e₁ ▷ʰ ↦ e₂ /⊢ wk) lemma₂
             (vs vz ,⇓ vz) ≡-refl
             T.eval-,
    where
    lemma₁ = prove
      (|wk| {Γ = ⌜ Γ ⌝▻} {σ = ⌜ σ₁ ⌝⋆} |∘ˢ| |wk| {σ = ⌜ σ₂ ⌝⋆})
      (|⌜ |wk| {Γ = Γ} {σ = σ₁} |∘ˢ| |wk| {σ = σ₂} ⌝⇒|)
      ≡-refl

    lemma₂ = prove
      (|⌜ liftʰ Σ₁ ⌝ʰ| |▷ʰ| |↦| |⌜ lift e₁ ⌝| |▷ʰ|
       |↦| |⌜ lift e₂ ⌝| |/⊢| |wk| {σ = ⌜ σ₁ ⌝⋆})
      (|⌜ liftʰ Σ₁ |▷ʰ| |↦| lift e₁ |▷ʰ|
       |↦| lift e₂ |/⊢| |wk| {σ = σ₁} ⌝ʰ|)
      ≡-refl

  complete' Γ Σ₁ (uncurry e) S.eval-uncurry _ _ =
    ∃thunked Γ ≡-refl idˢ (⌜id⌝-lemma Γ) Σ₁ ≡-refl
             (uncurry⇓ e) ≡-refl
             T.eval-uncurry

  -- Variables.

  complete' Γ Σ₁ (var x) d de eq =
    complete∋ Γ Σ₁ x d de eq

  -- Primitives.

  complete' Γ Σ₁ (return e) d de eq with complete' Γ Σ₁ e d de eq
  ... | ∃thunked Δ Δ-eq ρ' ρ-eq Σ' Σ-eq v' v-eq d' =
    ∃thunked Δ Δ-eq ρ' ρ-eq Σ' Σ-eq (return⋆⇓ v') v-eq
             (T.eval-return d')

  complete' Γ Σ₁ (√ e) d de eq with complete' Γ Σ₁ e d de eq
  ... | ∃thunked Δ Δ-eq ρ' ρ-eq Σ' Σ-eq (return⋆⇓ v') v-eq d' =
    ∃thunked Δ Δ-eq ρ' ρ-eq Σ' Σ-eq (return⋆⇓ v') v-eq
             (T.eval-√ d')

  complete' Γ Σ₁ (pay m₂ e) d de eq with complete' Γ Σ₁ e d de eq
  ... | ∃thunked Δ Δ-eq ρ' ρ-eq Σ' Σ-eq (return⋆⇓ v') v-eq d' =
    ∃thunked Δ Δ-eq ρ' ρ-eq Σ' Σ-eq (return⋆⇓ (return⋆⇓ v')) v-eq
             (T.eval-pay d')

  complete' Γ Σ₁ (e₁ >>= e₂) (S.eval-· d₁ d₂) [0]             ()
  complete' Γ Σ₁ (e₁ >>= e₂) (S.eval-· d₁ d₂) ([suc] _)       ()
  complete' Γ Σ₁ (e₁ >>= e₂) (S.eval-· d₁ d₂) (de₁ [max] de₂) eq
    with complete' Γ Σ₁ e₂ d₁ de₁ (drop-[max]d eq)
  ... | ∃thunked Δ ≡-refl ρ₁' ≡-refl Σ₂' ≡-refl v₂' ≡-refl d₁'
      with complete•
             Δ (Σ₂' ▷ʰ ↦ⁿ e₁ /⊢ ρ₁') v₂' (cast• d₂ lemma₁ ≡-refl)
             de₂ (cast•Eq depth• d₂ lemma₁ ≡-refl (drop-d[max] eq))
    where
    lemma₁ = prove
      (|⌜ liftʰ Σ₂' ⌝ʰ| |▷ʰ| |↦| |⌜ lift e₁ ⌝| |/⊢| |⌜ lift⇒ ρ₁' ⌝⇒|)
      (|⌜ liftʰ Σ₂' |▷ʰ| |↦ⁿ| lift e₁ |/⊢| lift⇒ ρ₁' ⌝ʰ|)
      ≡-refl

  ...   | ∃thunked• X ≡-refl ρ₂' ≡-refl Σ₃' ≡-refl
                    (return⋆⇓ v') ≡-refl d₂' =
    ∃thunked X ≡-refl (wk⇒ ρ₁' ∘ˢ ρ₂') lemma₂ Σ₃' ≡-refl
               (return⋆⇓ v') ≡-refl (T.eval->>= d₁' d₂')
    where
    lemma₂ = prove
      (|wk⇒| |⌜ lift⇒ ρ₁' ⌝⇒| |∘ˢ| |⌜ lift⇒ ρ₂' ⌝⇒|)
      (|⌜ |wk⇒| (lift⇒ ρ₁') |∘ˢ| lift⇒ ρ₂' ⌝⇒|)
      ≡-refl

  complete' Γ Σ₁ (gwait {σ = τ} {n₁ = m₁} C m₂ e) {v = v} d de eq
    with complete' Γ Σ₁ e {v = v'} (cast d allEq) de
                   (castEq depth d allEq eq)
    where
    lem = erase-lemma {m = m₁} {n = m₂ + m₁} {τ = τ} C
    v'  = ≡-subst (_⊢⇓_ _) (≡-sym lem) v

    allEq : AllEqual ⌜ Σ₁ ⌝ʰ _ _ (≡-subst (_⊢_ ⌜ Γ ⌝▻) lem ⌜ e ⌝) v
                     ⌜ Σ₁ ⌝ʰ _ _ ⌜ e ⌝ v'
    allEq = record
      { Γ-eq  = ≡-refl
      ; τ-eq  = ≡-sym lem
      ; Σ₁-eq = ≅-refl
      ; t-eq  = ≡-subst-removable (_⊢_ _) lem ⌜ e ⌝
      ; Δ-eq  = ≡-refl
      ; ρ-eq  = ≅-refl
      ; Σ₂-eq = ≅-refl
      ; v-eq  = ≅-sym (≡-subst-removable (_⊢⇓_ _) (≡-sym lem) v)
      }
  ... | ∃thunked Δ ≡-refl ρ' ≡-refl Σ' ≡-refl v' v-eq d' =
        ∃thunked Δ ≡-refl ρ' ≡-refl Σ' ≡-refl
                 (gwait⇓ C m₂ v') lemma
                 (T.eval-gwait {m₂ = m₂} {C = C} d')
    where
    lem'  = erase-lemma {m = m₁} {n = m₂ + m₁} {τ = τ} C
    cast₁ = ≡-subst (_⊢⇓_ ⌜ Δ ⌝▻) lem'
    cast₂ = ≡-subst (_⊢⇓_ ⌜ Δ ⌝▻) (≡-sym lem')
    lemma = begin
      v                    ≡⟨ ≅-to-≡ (≅-sym (≡-subst-removable (_⊢⇓_ _) (≡-sym lem') v)
                                        ⟨ ≅-trans ⟩
                                      ≅-sym (≡-subst-removable (_⊢⇓_ _) lem' _)) ⟩
      cast₁ (cast₂ v)      ≡⟨ ≡-cong cast₁ v-eq ⟩
      cast₁ ⌜ v' ⌝⇓        ≡⟨ prove (|≡-subst⇓| lem' |⌜ lift v' ⌝⇓|)
                                    |⌜ |gwait⇓| C m₂ (lift v') ⌝⇓|
                                    ≡-refl ⟩
      ⌜ gwait⇓ C m₂ v' ⌝⇓  ∎

  complete' Γ Σ₁ (gpay {σ = τ} {n₁ = m₁} C m₂ e) {v = v} d de eq
    with complete' Γ Σ₁ e {v = v'} (cast d allEq) de
                   (castEq depth d allEq eq)
    where
    lem = erase-lemma {m = m₁} {n = m₁ ∸ m₂} {τ = τ} C
    v'  = ≡-subst (_⊢⇓_ _) (≡-sym lem) v

    allEq : AllEqual ⌜ Σ₁ ⌝ʰ _ _ (≡-subst (_⊢_ ⌜ Γ ⌝▻) lem ⌜ e ⌝) v
                     ⌜ Σ₁ ⌝ʰ _ _ ⌜ e ⌝ v'
    allEq = record
      { Γ-eq  = ≡-refl
      ; τ-eq  = ≡-sym lem
      ; Σ₁-eq = ≅-refl
      ; t-eq  = ≡-subst-removable (_⊢_ _) lem ⌜ e ⌝
      ; Δ-eq  = ≡-refl
      ; ρ-eq  = ≅-refl
      ; Σ₂-eq = ≅-refl
      ; v-eq  = ≅-sym (≡-subst-removable (_⊢⇓_ _) (≡-sym lem) v)
      }
  ... | ∃thunked Δ ≡-refl ρ' ≡-refl Σ' ≡-refl v' v-eq d' =
    ∃thunked (GPay.Δ r) Δ-eq
             (ρ' ∘ˢ GPay.ρ r) (∘-ρ-lemma rs ρ')
             (GPay.Σ₂ r) (GPaySound.Σ₂-eq rs)
             (return⋆⇓ (GPay.v₂ r)) lemma
             (T.eval-gpay {C = C} d')
    where
    lem' = erase-lemma {m = m₁} {n = m₁ ∸ m₂} {τ = τ} C
    r    = gpayResult C m₂ Σ' v'
    rs   = gpayResult-sound C m₂ Σ' v'
    Δ-eq = GPaySound.Δ-eq rs

    cast₁ = ≡-subst (\Δ -> Δ ⊢⇓ ⌜ C [ Thunk (m₁ ∸ m₂) τ ] ⌝⋆) Δ-eq
    cast₂ = ≡-subst (_⊢⇓_ ⌜ Δ ⌝▻) lem'
    cast₃ = ≡-subst (_⊢⇓_ ⌜ Δ ⌝▻) (≡-sym lem')

    lemma = begin
      cast₁ v                  ≡⟨ ≡-cong cast₁ (≅-to-≡
                                    (≅-sym (≡-subst-removable (_⊢⇓_ _) (≡-sym lem') v)
                                       ⟨ ≅-trans ⟩
                                     ≅-sym (≡-subst-removable (_⊢⇓_ _) lem' (cast₃ v)))) ⟩
      cast₁ (cast₂ (cast₃ v))  ≡⟨ ≡-cong (cast₁ ∘ cast₂) v-eq ⟩
      cast₁ (cast₂ ⌜ v' ⌝⇓)    ≡⟨ GPaySound.v₂-eq rs ⟩
      ⌜ GPay.v₂ r ⌝⇓           ∎

  -- Fixpoints.

  complete' Γ Σ₁ (fix e) (S.eval-fix d) [0] ()
  complete' Γ Σ₁ (fix e) (S.eval-fix d) (_ [max] _) ()
  complete' Γ Σ₁ (fix e) (S.eval-fix d) ([suc] de) eq
    with complete' (Γ ▻ _) (Σ₁ ▷ʰ ↦ fix e) e d de (drop-[suc] eq)
  ... | ∃thunked Δ ≡-refl ρ' ≡-refl Σ' ≡-refl (return⋆⇓ v') ≡-refl d' =
    ∃thunked Δ ≡-refl (wk ∘ˢ ρ') lemma Σ' ≡-refl
             (return⋆⇓ v') ≡-refl (T.eval-fix d')
    where
    lemma = prove
      (|wk| |∘ˢ| |⌜ lift⇒ ρ' ⌝⇒|)
      (|⌜ |wk| |∘ˢ| lift⇒ ρ' ⌝⇒|)
      ≡-refl

  -- Applications.

  complete' Γ Σ₁ (e₁ · e₂) (S.eval-· d₁ d₂) [0]             ()
  complete' Γ Σ₁ (e₁ · e₂) (S.eval-· d₁ d₂) ([suc] _)       ()
  complete' Γ Σ₁ (e₁ · e₂) (S.eval-· d₁ d₂) (de₁ [max] de₂) eq
    with complete' Γ Σ₁ e₁ d₁ de₁ (drop-[max]d eq)
  ... | ∃thunked Δ ≡-refl ρ₁' ≡-refl Σ₂' ≡-refl v₁' ≡-refl d₁'
      with complete•
             Δ (Σ₂' ▷ʰ ↦ e₂ /⊢ ρ₁') v₁' (cast• d₂ lemma₁ ≡-refl)
             de₂ (cast•Eq depth• d₂ lemma₁ ≡-refl (drop-d[max] eq))
    where
    lemma₁ = prove
      (|⌜ liftʰ Σ₂' ⌝ʰ| |▷ʰ| |↦| |⌜ lift e₂ ⌝| |/⊢| |⌜ lift⇒ ρ₁' ⌝⇒|)
      (|⌜ liftʰ Σ₂' |▷ʰ| |↦| lift e₂ |/⊢| lift⇒ ρ₁' ⌝ʰ|)
      ≡-refl

  ...   | ∃thunked• X ≡-refl ρ₂' ≡-refl Σ₃' ≡-refl v' ≡-refl d₂' =
    ∃thunked X ≡-refl (wk⇒ ρ₁' ∘ˢ ρ₂') lemma₂ Σ₃' ≡-refl v' ≡-refl
             (T.eval-· d₁' d₂')
    where
    lemma₂ = prove
      (|wk⇒| |⌜ lift⇒ ρ₁' ⌝⇒| |∘ˢ| |⌜ lift⇒ ρ₂' ⌝⇒|)
      (|⌜ |wk⇒| (lift⇒ ρ₁') |∘ˢ| lift⇒ ρ₂' ⌝⇒|)
      ≡-refl

  complete• :  forall Γ {Δ σ τ} Σ₁ {ρ Σ₂ n}
               (v₀ : Γ ⊢⇓ σ → τ) {v : Δ ⊢⇓ ⌜ τ ⌝⋆}
            -> (deriv : ⌜ Σ₁ ⌝ʰ ∣ ⌜ v₀ ⌝⇓ •⇓ ρ ⋉ Σ₂ ∣ v √ n)
            -> (d : Depth) -> depth• deriv ≡ d
            -> ∃Thunked• Σ₁ v₀ ρ Σ₂ v n
  complete• Γ Σ₁ s⇓ S.eval-·-s _ _ =
    ∃thunked• (Γ ▻ Nat) ≡-refl idˢ (⌜id⌝-lemma (Γ ▻ Nat)) Σ₁ ≡-refl
              (s₁⇓ vz) ≡-refl T.eval-·-s

  complete• Γ {σ = σ} Σ₁ (λ⇓ e) (S.eval-·-λ d) [0]         ()
  complete• Γ {σ = σ} Σ₁ (λ⇓ e) (S.eval-·-λ d) (_ [max] _) ()
  complete• Γ {σ = σ} Σ₁ (λ⇓ e) (S.eval-·-λ d) ([suc] de)  eq
    with complete' (Γ ▻ σ) Σ₁ e d de (drop-[suc] eq)
  ... | ∃thunked Δ ≡-refl ρ' ≡-refl Σ₂' ≡-refl
                 (return⋆⇓ v') ≡-refl d' =
    ∃thunked• Δ ≡-refl ρ' ≡-refl Σ₂' ≡-refl (return⋆⇓ v') ≡-refl
              (T.eval-·-λ d')

  complete• Γ {σ = σ₁} Σ₁ (λ²⇓ e₁) S.eval-·-λ² _ _ =
    ∃thunked• (Γ ▻ σ₁) ≡-refl idˢ (⌜id⌝-lemma (Γ ▻ σ₁)) Σ₁ ≡-refl
              (λ⇓ e₁) ≡-refl T.eval-·-λ²

  -- Natrec zero.

  complete• Γ Σ₁ (natrec⇓ x₁¹ e₁²) (S.eval-·-natrec-zero d₁ d₂)
            [0] ()
  complete• Γ Σ₁ (natrec⇓ x₁¹ e₁²) (S.eval-·-natrec-zero d₁ d₂)
            ([suc] _) ()
  complete• Γ Σ₁ (natrec⇓ x₁¹ e₁²) (S.eval-·-natrec-zero d₁ d₂)
            (de₁ [max] de₂) eq
    with complete' (Γ ▻ Nat) Σ₁ (var vz) d₁ de₁ (drop-[max]d eq)
  ... | ∃thunked Δ ≡-refl ρ₁' ≡-refl Σ₂' ≡-refl (s₁⇓ _) ()     d₁'
  ... | ∃thunked Δ ≡-refl ρ₁' ≡-refl Σ₂' ≡-refl z⇓      ≡-refl d₁'
    with complete'
           Δ Σ₂' (var x₁¹ /⊢ wk ∘ˢ ρ₁')
           (cast d₂ allEq) de₂
           (castEq depth d₂ allEq (drop-d[max] eq))
    where
    lemma₁ = prove
      (|var| |⌜ lift x₁¹ ⌝∋| |/⊢| |wk| |∘ˢ| |⌜ lift⇒ ρ₁' ⌝⇒|)
      (|⌜ |var| (lift x₁¹) |/⊢| |wk| |∘ˢ| lift⇒ ρ₁' ⌝|)
      ≡-refl

    allEq = termsEqual lemma₁

  ...   | ∃thunked X ≡-refl ρ₂' ≡-refl Σ₃' ≡-refl
                   (return⋆⇓ v') ≡-refl d₂' =
    ∃thunked• X ≡-refl (ρ₁' ∘ˢ ρ₂') lemma₂ Σ₃' ≡-refl
              (return⋆⇓ v') ≡-refl (T.eval-·-natrec-zero d₁' d₂')
    where
    lemma₂ = prove
      (|⌜ lift⇒ ρ₁' ⌝⇒| |∘ˢ| |⌜ lift⇒ ρ₂' ⌝⇒|)
      (|⌜ lift⇒ ρ₁' |∘ˢ| lift⇒ ρ₂' ⌝⇒|)
      ≡-refl

  -- Natrec suc.

  complete• Γ Σ₁ (natrec⇓ x₁¹ e₁²) (S.eval-·-natrec-suc d₁ d₂)
            [0] ()
  complete• Γ Σ₁ (natrec⇓ x₁¹ e₁²) (S.eval-·-natrec-suc d₁ d₂)
            ([suc] _) ()
  complete• Γ Σ₁ (natrec⇓ x₁¹ e₁²) (S.eval-·-natrec-suc d₁ d₂)
            (de₁ [max] de₂) eq
    with complete' (Γ ▻ Nat) Σ₁ (var vz) d₁ de₁ (drop-[max]d eq)
  ... | ∃thunked Δ ≡-refl ρ₁' ≡-refl Σ₂' ≡-refl z⇓        ()     d₁'
  ... | ∃thunked Δ ≡-refl ρ₁' ≡-refl Σ₂' ≡-refl (s₁⇓ x₂') ≡-refl d₁'
    with complete'
           (Δ ▻ _) Σ₂'' e₁²' (cast d₂ allEq)
           de₂ (castEq depth d₂ allEq (drop-d[max] eq))
    where
    Σ₂'' = Σ₂' ▷ʰ ↦ (natrec (var x₁¹) e₁² /⊢ (wk ∘ˢ ρ₁')) · var x₂'
    e₁²' = e₁² /⊢ (wk ∘ˢ ρ₁') ↑ ↑ /⊢ sub x₂' ↑

    lemma₁ = prove
      (|⌜ liftʰ Σ₂' ⌝ʰ| |▷ʰ|
       |↦| ((|natrec| (|var| |⌜ lift x₁¹ ⌝∋|) |⌜ lift e₁² ⌝| |/⊢|
             |wk| |∘ˢ| |⌜ lift⇒ ρ₁' ⌝⇒|) |·|
            |var| |⌜ lift x₂' ⌝∋|))
      (|⌜ liftʰ Σ₂' |▷ʰ|
       |↦| (|natrec| (|var| (lift x₁¹)) (lift e₁²) |/⊢|
            |wk| |∘ˢ| lift⇒ ρ₁') |·|
           |var| (lift x₂') ⌝ʰ|)
      ≡-refl

    lemma₂ = prove
      (|⌜ lift e₁² ⌝| |/⊢| (|wk| |∘ˢ| |⌜ lift⇒ ρ₁' ⌝⇒|) |↑| |↑| |/⊢|
       |sub| |⌜ lift x₂' ⌝∋| |↑|)
      (|⌜ lift e₁² |/⊢| (|wk| |∘ˢ| lift⇒ ρ₁') |↑| |↑| |/⊢|
       |sub| (lift x₂') |↑| ⌝|)
      ≡-refl

    allEq = inputsEqual lemma₁ lemma₂

  ...   | ∃thunked X ≡-refl ρ₂' ≡-refl Σ₃' ≡-refl
                   (return⋆⇓ v') ≡-refl d₂' =
    ∃thunked• X ≡-refl (wk⇒ ρ₁' ∘ˢ ρ₂') lemma₃ Σ₃' ≡-refl
              (return⋆⇓ v') ≡-refl (T.eval-·-natrec-suc d₁' d₂')
    where
    lemma₃ = prove
      (|wk⇒| |⌜ lift⇒ ρ₁' ⌝⇒| |∘ˢ| |⌜ lift⇒ ρ₂' ⌝⇒|)
      (|⌜ |wk⇒| (lift⇒ ρ₁') |∘ˢ| lift⇒ ρ₂' ⌝⇒|)
      ≡-refl

  -- Uncurry.

  complete• Γ Σ₁ (uncurry⇓ e₁) (S.eval-·-uncurry d₁ d₂) [0]       ()
  complete• Γ Σ₁ (uncurry⇓ e₁) (S.eval-·-uncurry d₁ d₂) ([suc] _) ()
  complete• Γ Σ₁ (uncurry⇓ e₁) (S.eval-·-uncurry d₁ d₂)
            (de₁ [max] de₂) eq
    with complete' (Γ ▻ _) Σ₁ (var vz) d₁ de₁ (drop-[max]d eq)
  ... | ∃thunked Δ ≡-refl ρ₁' ≡-refl Σ₂' ≡-refl
                 (x₂' ,⇓ y₂') ≡-refl d₁'
    with complete'
           Δ Σ₂' e₁' (cast d₂ allEq)
           de₂ (castEq depth d₂ allEq (drop-d[max] eq))
    where
    e₁' = e₁ /⊢ (wk ∘ˢ ρ₁') ↑ ↑ /⊢ sub x₂' ↑ /⊢ sub y₂'

    lemma₁ = prove
      (|⌜ lift e₁ ⌝| |/⊢| (|wk| |∘ˢ| |⌜ lift⇒ ρ₁' ⌝⇒|) |↑| |↑| |/⊢|
       |sub| |⌜ lift x₂' ⌝∋| |↑| |/⊢| |sub| |⌜ lift y₂' ⌝∋|)
      (|⌜ lift e₁ |/⊢| (|wk| |∘ˢ| lift⇒ ρ₁') |↑| |↑| |/⊢|
       |sub| (lift x₂') |↑| |/⊢| |sub| (lift y₂') ⌝|)
      ≡-refl

    allEq = termsEqual lemma₁

  ...   | ∃thunked X ≡-refl ρ₂' ≡-refl Σ₃' ≡-refl
                   (return⋆⇓ v') ≡-refl d₂' =
    ∃thunked• X ≡-refl (ρ₁' ∘ˢ ρ₂') lemma₂ Σ₃' ≡-refl
              (return⋆⇓ v') ≡-refl (T.eval-·-uncurry d₁' d₂')
    where
    lemma₂ = prove
      (|⌜ lift⇒ ρ₁' ⌝⇒| |∘ˢ| |⌜ lift⇒ ρ₂' ⌝⇒|)
      (|⌜ lift⇒ ρ₁' |∘ˢ| lift⇒ ρ₂' ⌝⇒|)
      ≡-refl
